/* $Header */

/*
 * Copyright (C) 1994 by Premia Corporation.  All rights reserved.
 *
 */

#include <windows.h>
#include <string.h>
#include "exports.h"
#include "cwstart.h"

#define MovToColumn(x)		MovRight((x) - BufQCurrentCol())

/* 
 * These macros, from cwstart\util.c, are for _do_block().
 */

#define BLK_REPLACE       1 // Replace sel with string pointed at by arg
#define BLK_REPLACEFREE   2	// Same as BKL_REPLACE but free the ptr when done.
#define BLK_IGNORE        3 // Ignore current line.
#define BLK_DELETE        4 // Delete the current line.

static BOOL hasTabs(void);


LIBMAIN

void DLL
_init()
{
    LibExport( "void ColCenter BOOL");
    LibExport( "void ColRightJustify BOOL");
    LibExport( "void ColLeftJustify BOOL");
    
    /*  These entry points aren't meant to be public.
     *  They are callbacks used by the column stuff.
     */
    LibExport( "int _col_center long");
    LibExport( "int _col_right_justify long");
    LibExport( "int _col_left_justify long");
	
    /*  This function is defined in cwstart\util.c.
	 *  It performs a given operation on each line of
	 *  a column selection.  It needs to be exported
	 *  here because we call it via LibFunctionExec.
	 */
    LibExport( "void cwstart:_do_block LPSTR");
}
    

/*
 *  These functions justify text.  They can left-justify, right-justify,
 *  or center.  They can operate on line, stream, or column selections.
 *  They can optionally handle selections containing tabs.
 *
 *  In the case of a stream or line selection, each line that is part
 *  of the selection gets justified.  The left edge is always column 1.
 *  The right edge is the visible margin mark, if there is one.  Otherwise,
 *  the wrap margin is used.
 *
 *  In the case of a column selection, only the text in the selection
 *  gets justified.  The left and right bounds of the selection
 *  are used for the reference points.
 *
 *  If no text is selected, the entire current line gets adjusted.
 *
 *  There are three entry points: ColCenter, ColLeftJustify, and
 *  ColRightJustify.  Each takes a single boolean argument.  If this
 *  argument is TRUE, then the function will attempt to handle tabs
 *  correctly.  It will navigate the line by reckoning columns rather
 *  than characters.  If the argument is FALSE, the routines assume
 *  that each character takes up exactly one column.  This assumption
 *  allows some greatly time-saving optimizations.  If no argument
 *  is given, tab-handling is disabled.
 */	
 
void DLL
ColCenter(BOOL dealWithTabs)
{
    LPSelection selp;
    MarkID marktype;
    BOOL wasMark = TRUE;
    long rightMargin, center;
    long newLeft;
    LPSTR buf;
    LPSTR rStr, lStr;
    long strLen;
        
    if (ArgCount() < 1)
        dealWithTabs = 2;	// Autodetect
        
    SysWaitStart();		
    
    if (!(selp = MemAlloc(sizeof(Selection))))
        return;
        
    if (!(marktype = MarkQSelType()))
    {
        wasMark = FALSE;
        MarkBeginSel(SELECTION_LINE);
    }
    MarkNormalizeSel();
    marktype = MarkGetSelInfo(selp, TRUE);
    MarkSavePos();
    /* center the selection relative to its horizontal extent	*/
    if (marktype == SELECTION_COLUMN)
    {
		if (dealWithTabs == 2)
			dealWithTabs = hasTabs();
        if (!dealWithTabs)
			LibFunctionExec("_do_block _col_center");
        else
        {
            LPSTR buf, rStr;
            long width, leftPad, rightPad, strLen;
            long i, eol;
            
            width = (selp->end.column - selp->begin.column) + 1;
            MovToLine(selp->begin.line);
            
            while (BufQCurrentLine() <= selp->end.line)
            {
                if (BufQPastEOF())
                    break;
                
                MovEOL();
                if ((eol = BufQCurrentCol()) > 0)
                    --eol;
                if (selp->begin.column <= eol)
                {
                    MovToColumn(selp->begin.column);	
                    MarkBeginSel(SELECTION_INCLUSIVE);
                    MovToColumn(min(selp->end.column, eol));
                    buf = MarkGetSelStr(NULL, NULL);	
                    BufDelSelection();

                    rStr = StrTrim(buf, " \t");
                    StrFree(buf);
                    buf = StrLtrim(rStr, " \t");
                    strLen = strlen(buf);
                    leftPad = (width - strLen) / 2;
                    rightPad = width - (leftPad + strLen);
                    
                    for (i = 0; i < leftPad; ++i)
                        BufInsertStr(" ");
                    BufInsertStr(buf);
                    for (i = 0; i < rightPad; ++i)
                        BufInsertStr(" ");
                    
                    StrFree(rStr);
                    StrFree(buf);
                }
                MovDown(1);
            }
        }
        MemFree(selp);
        MarkRestorePos();
    }
    /* center the selection relative to the right margin mark
     * if there is one.  Otherwise, use the wrap margin.
     */
    else
    {
        if (!(rightMargin = WinVisibleMarginColumn(-1)))
            rightMargin = WrapSetRightMargin(-1);
    
        center = rightMargin / 2;
    
        MovToLine(selp->begin.line);
        while (BufQCurrentLine() <= selp->end.line)
        {
            if (BufQPastEOF())
                break;
            MovHome();
            buf = BufReadStr(0x7FFFFFFF);
            rStr = StrTrim(buf, " \t");
            lStr = StrLtrim(rStr, " \t");
            StrFree(buf);
            StrFree(rStr);
            buf = lStr;
            strLen = strlen(buf);
            newLeft = center - strLen / 2;
            BufDelToEOL();
            MovRight(newLeft);
            BufInsertStr(buf);
            MemFree(buf);
            if (!wasMark)
                MarkRemoveSel();
            MovDown(1);
        }	
        MemFree(selp);
        MarkRestorePos();
    }	
    if (!wasMark)			/* make sure selection gets removed		*/
        MarkRemoveSel();
    SysWaitEnd();		
}



void DLL
ColRightJustify(BOOL dealWithTabs)
{
    LPSelection selp;
    MarkID marktype;
    BOOL wasMark = TRUE;
    long rightMargin;
    long newLeft;
    LPSTR buf;
    LPSTR rStr, lStr;
    long strLen;
            
    
    if (!(selp = MemAlloc(sizeof(Selection))))
        return;
    
    if (ArgCount() < 1)
        dealWithTabs = 2;
        
    SysWaitStart();	
        
    if (!(marktype = MarkQSelType()))
    {
        wasMark = FALSE;
        MarkBeginSel(SELECTION_LINE);
    }
    MarkNormalizeSel();
    marktype = MarkGetSelInfo(selp, TRUE);
    MarkSavePos();
    /* right-justify the selection relative to its horizontal extent	*/
    if (marktype == SELECTION_COLUMN)
    {
		if (dealWithTabs == 2)
			dealWithTabs = hasTabs();
        if (!dealWithTabs)
            LibFunctionExec("_do_block _col_right_justify");
        else
        {
            LPSTR buf, rStr;
            long width, leftPad, strLen;
            long i, eol;
            
            width = (selp->end.column - selp->begin.column) + 1;
            MovToLine(selp->begin.line);
            
            while (BufQCurrentLine() <= selp->end.line)
            {
                if (BufQPastEOF())
                    break;
                
                MovEOL();
                if ((eol = BufQCurrentCol()) > 0)
                    --eol;
                MovToColumn(selp->begin.column);	
                MarkBeginSel(SELECTION_INCLUSIVE);
                MovToColumn(min(selp->end.column, eol));
                buf = MarkGetSelStr(NULL, NULL);	
                BufDelSelection();

                rStr = StrTrim(buf, " \t");
                StrFree(buf);
                buf = StrLtrim(rStr, " \t");
                strLen = strlen(buf);
                leftPad = width - strLen;
                
                for (i = 0; i < leftPad; ++i)
                    BufInsertStr(" ");
                BufInsertStr(buf);    
                    
                MovDown(1);    
                    StrFree(rStr);
                StrFree(buf);    
            }
        }    
        MemFree(selp);
        MarkRestorePos();    
        }
    /* right-justify the selection relative to the right margin mark    
     * if there is one.  Otherwise, use the wrap margin.
     */    
    else
    {
        if (!(rightMargin = WinVisibleMarginColumn(-1)))
            rightMargin = WrapSetRightMargin(-1);
    
        MovToLine(selp->begin.line);
        while (BufQCurrentLine() <= selp->end.line)
        {
            if (BufQPastEOF())
                break;
            MovHome();
            buf = BufReadStr(0x7FFFFFFF);
            rStr = StrTrim(buf, " \t");
            lStr = StrLtrim(rStr, " \t");
            StrFree(buf);
            StrFree(rStr);
            buf = lStr;
            strLen = strlen(buf);
            newLeft = rightMargin - strLen;
            BufDelToEOL();
            MovRight(newLeft);
            BufInsertStr(buf);
            MemFree(buf);
            if (!wasMark)
                MarkRemoveSel();
            MovDown(1);
        }	
        MemFree(selp);
        MarkRestorePos();
    }	
    if (!wasMark)			/* make sure selection gets removed		*/
        MarkRemoveSel();
    SysWaitEnd();
}

void DLL
ColLeftJustify(BOOL dealWithTabs)
{
    LPSelection selp;
    MarkID marktype;
    BOOL wasMark = TRUE;
    long rightMargin;
    LPSTR buf;
    LPSTR rStr, lStr;
    long strLen;
            
    if (ArgCount() < 1)
        dealWithTabs = 2;
    
    if (!(selp = MemAlloc(sizeof(Selection))))
        return;
    
    SysWaitStart();	
        
    if (!(marktype = MarkQSelType()))
    {
        wasMark = FALSE;
        MarkBeginSel(SELECTION_LINE);
    }
    MarkNormalizeSel();
    marktype = MarkGetSelInfo(selp, TRUE);
    MarkSavePos();
    /* left-justify the selection relative to its horizontal extent	*/
    if (marktype == SELECTION_COLUMN)
    {
        if (dealWithTabs == 2)
			dealWithTabs = hasTabs();
        if (!dealWithTabs)
            LibFunctionExec("_do_block _col_left_justify");
        else
        {
            LPSTR buf, rStr;
            long width, rightPad, strLen;
            long i, eol;
            
            width = (selp->end.column - selp->begin.column) + 1;
            MovToLine(selp->begin.line);
            
            while (BufQCurrentLine() <= selp->end.line)
            {
                if (BufQPastEOF())
                    break;
                
                MovEOL();
                if ((eol = BufQCurrentCol()) > 0)
                    --eol;
                MovToColumn(selp->begin.column);	
                MarkBeginSel(SELECTION_INCLUSIVE);
                MovToColumn(min(selp->end.column, eol));
                buf = MarkGetSelStr(NULL, NULL);	
                BufDelSelection();

                rStr = StrTrim(buf, " \t");
                StrFree(buf);
                buf = StrLtrim(rStr, " \t");
                strLen = strlen(buf);
                rightPad = width - strLen;
                
                BufInsertStr(buf);
                for (i = 0; i < rightPad; ++i)
                    BufInsertStr(" ");
                
                MovDown(1);
                StrFree(rStr);
                StrFree(buf);
            }
        }
        MemFree(selp);
        MarkRestorePos();
    }
    /* left-justify the selection against the column 1.
     */
    else
    {
        if (!(rightMargin = WinVisibleMarginColumn(-1)))
            rightMargin = WrapSetRightMargin(-1);
    
        MovToLine(selp->begin.line);
        while (BufQCurrentLine() <= selp->end.line)
        {
            if (BufQPastEOF())
                break;
            MovHome();
            buf = BufReadStr(0x7FFFFFFF);
            rStr = StrTrim(buf, " \t");
            lStr = StrLtrim(rStr, " \t");
            StrFree(buf);
            StrFree(rStr);
            buf = lStr;
            strLen = strlen(buf);
            BufDelToEOL();
            BufInsertStr(buf);
            MemFree(buf);
            if (!wasMark)
                MarkRemoveSel();
            MovDown(1);
        }	
        MemFree(selp);
        MarkRestorePos();
    }	
    if (!wasMark)			/* make sure selection gets removed		*/
        MarkRemoveSel();
    SysWaitEnd();
}

/*
 *  These entry points aren't meant to be public.
 *  They are callbacks used by the column stuff.
 *
 *  If you really, really want to use them, look
 *  at the way _do_block does the callback in
 *  util.c.  Basically, you can pass the address
 *  of a LPSTR that points to a null-terminated.
 *  string.  This function does not free that pointer.
 *  It resets the pointer to look at an allocated
 *  string, which the caller must free.
 */
 
int DLL
_col_center(LPVOID XFAR *blockBufPtr)
{
    LPSTR buf, rStr, blockPtr;               
    long width, leftPad, rightPad, strLen;   
    long i;                                  
    static long top, left, bot, right;       
    
    MarkCoordinates(&top, &left, &bot, &right);
    
    width = (right - left) + 1;
    rStr = StrTrim(*blockBufPtr, " \t");
    buf = StrLtrim(rStr, " \t");
        
    strLen = strlen(buf);
    leftPad = (width - strLen) / 2;
    rightPad = width - (leftPad + strLen);
    
	if (strLen >= width)
	{
		*blockBufPtr = StrNNew(buf, (int)width);
	    return(BLK_REPLACEFREE);
	}
    *blockBufPtr = blockPtr = MemAlloc((int)(width + 1));
    
    for (i = 0; i < leftPad && i < width; ++i)
        *blockPtr++ = ' ';
    for (i = 0; i < strLen && i + leftPad < width; ++i)
        *blockPtr++ = buf[i];
    for (i = 0; i < rightPad && i + leftPad + strLen < width; ++i)
        *blockPtr++ = ' ';
    *blockPtr = 0;
    
    StrFree(rStr);
    StrFree(buf);
    
    return(BLK_REPLACEFREE);
}

int DLL
_col_right_justify(LPVOID XFAR *blockBufPtr)
{
    LPSTR buf, rStr, blockPtr;
    long width, leftPad, strLen;
    long i;
    static long top, left, bot, right;
    
    MarkCoordinates(&top, &left, &bot, &right);
    
    width = (right - left) + 1;
    rStr = StrTrim(*blockBufPtr, " \t");
    buf = StrLtrim(rStr, " \t");
	
    strLen = strlen(buf);
	if (strLen >= width)
	{
		*blockBufPtr = StrNNew(buf, (int)width);
	    return(BLK_REPLACEFREE);
	}
    leftPad = width - strLen;
    
    *blockBufPtr = blockPtr = MemAlloc((int)(width + 1));
    
    for (i = 0; i < leftPad && i < width; ++i)
        *blockPtr++ = ' ';
    for (i = 0; i < strLen && i + leftPad < width; ++i)
        *blockPtr++ = buf[i];
    *blockPtr = 0;
    
    StrFree(rStr);
    StrFree(buf);
    
    return(BLK_REPLACEFREE);
}

int DLL
_col_left_justify(LPVOID XFAR *blockBufPtr)
{
    LPSTR buf, rStr, blockPtr;
    long width, rightPad, strLen;
    long i;
    static long top, left, bot, right;
    
    MarkCoordinates(&top, &left, &bot, &right);
    
    width = (right - left) + 1;
    rStr = StrTrim(*blockBufPtr, " \t");
    buf = StrLtrim(rStr, " \t");
        
    strLen = strlen(buf);
    rightPad = width - strLen;
    
	if (strLen >= width)
	{
		*blockBufPtr = StrNNew(buf, (int)width);
	    return(BLK_REPLACEFREE);
	}
    *blockBufPtr = blockPtr = MemAlloc((int)(width + 1));
    
    for (i = 0; i < strLen && i < width; ++i)
        *blockPtr++ = buf[i];
    for (i = 0; i < rightPad && i + strLen < width; ++i)
        *blockPtr++ = ' ';
    *blockPtr = 0;
    
    StrFree(rStr);
    StrFree(buf);
    
    return(BLK_REPLACEFREE);
}


/*
 * Return 0 iff the current selection contains tabs or contains
 * virtual space associated with a preceding tab.
 */
static BOOL
hasTabs(void)
{
	LPSTR str, p;
	
	if (p = str = MarkGetSelStr(NULL, NULL))
	{
		while (*p)
		{
			if (*p == -1)
			{
				while (*p == -1)
					++p;
				if (*p && *p != '\n')
					break;	// return TRUE
			}
			if (*p == '\t')
				break;
			++p;
		}
	}
	StrFree(str);
	if (!*p)
		return(FALSE);
	else	
		return(TRUE);
}
